@@ -4,8 +4,10 @@
 #include "msg/Messenger.h"
 #include "ObjectCacher.h"
 #include "WritebackHandler.h"
+#include "common/errno.h"
 #include "common/perf_counters.h"
 
+#include <limits.h>
 
 /*** ObjectCacher::BufferHead ***/
 
@@ -27,7 +29,7 @@ ObjectCacher::BufferHead *ObjectCacher::Object::split(BufferHead *left, loff_t o
   right->last_write_tid = left->last_write_tid;
   right->set_state(left->get_state());
   right->snapc = left->snapc;
-  
+
   loff_t newleftlen = off - left->start();
   right->set_start(off);
   right->set_length(left->length() - newleftlen);
@@ -114,15 +116,16 @@ void ObjectCacher::Object::try_merge_bh(BufferHead *bh)
 	p->second->get_state() == bh->get_state()) {
       merge_left(p->second, bh);
       bh = p->second;
-    } else 
+    } else {
       p++;
+    }
   }
   // to the right?
   assert(p->second == bh);
   p++;
   if (p != data.end() &&
       p->second->start() == bh->end() &&
-      p->second->get_state() == bh->get_state()) 
+      p->second->get_state() == bh->get_state())
     merge_left(bh, p->second);
 }
 
@@ -131,15 +134,7 @@ void ObjectCacher::Object::try_merge_bh(BufferHead *bh)
  */
 bool ObjectCacher::Object::is_cached(loff_t cur, loff_t left)
 {
-  map<loff_t, BufferHead*>::iterator p = data.lower_bound(cur);
-  
-  if (p != data.begin() && 
-      (p == data.end() || p->first > cur)) {
-    p--;     // might overlap!
-    if (p->first + p->second->length() <= cur) 
-      p++;   // doesn't overlap.
-  }
-  
+  map<loff_t, BufferHead*>::iterator p = data_lower_bound(cur);
   while (left > 0) {
     if (p == data.end())
       return false;
@@ -168,7 +163,8 @@ bool ObjectCacher::Object::is_cached(loff_t cur, loff_t left)
 int ObjectCacher::Object::map_read(OSDRead *rd,
                                    map<loff_t, BufferHead*>& hits,
                                    map<loff_t, BufferHead*>& missing,
-                                   map<loff_t, BufferHead*>& rx)
+                                   map<loff_t, BufferHead*>& rx,
+				   map<loff_t, BufferHead*>& errors)
 {
   for (vector<ObjectExtent>::iterator ex_it = rd->extents.begin();
        ex_it != rd->extents.end();
@@ -177,21 +173,13 @@ int ObjectCacher::Object::map_read(OSDRead *rd,
     if (ex_it->oid != oid.oid) continue;
     
     ldout(oc->cct, 10) << "map_read " << ex_it->oid 
-             << " " << ex_it->offset << "~" << ex_it->length << dendl;
-    
-    map<loff_t, BufferHead*>::iterator p = data.lower_bound(ex_it->offset);
-    // p->first >= start
+		       << " " << ex_it->offset << "~" << ex_it->length
+		       << dendl;
     
     loff_t cur = ex_it->offset;
     loff_t left = ex_it->length;
-    
-    if (p != data.begin() && 
-        (p == data.end() || p->first > cur)) {
-      p--;     // might overlap!
-      if (p->first + p->second->length() <= cur) 
-        p++;   // doesn't overlap.
-    }
-    
+
+    map<loff_t, BufferHead*>::iterator p = data_lower_bound(ex_it->offset);
     while (left > 0) {
       // at end?
       if (p == data.end()) {
@@ -212,18 +200,21 @@ int ObjectCacher::Object::map_read(OSDRead *rd,
       if (p->first <= cur) {
         // have it (or part of it)
         BufferHead *e = p->second;
-        
+
         if (e->is_clean() ||
             e->is_dirty() ||
             e->is_tx()) {
           hits[cur] = e;     // readable!
           ldout(oc->cct, 20) << "map_read hit " << *e << dendl;
-        } 
-        else if (e->is_rx()) {
+        } else if (e->is_rx()) {
           rx[cur] = e;       // missing, not readable.
           ldout(oc->cct, 20) << "map_read rx " << *e << dendl;
-        }
-        else assert(0);
+        } else if (e->is_error()) {
+	  errors[cur] = e;
+	  ldout(oc->cct, 20) << "map_read error " << *e << dendl;
+	} else {
+	  assert(0);
+	}
         
         loff_t lenfromcur = MIN(e->end() - cur, left);
         cur += lenfromcur;
@@ -243,12 +234,12 @@ int ObjectCacher::Object::map_read(OSDRead *rd,
         left -= MIN(left, n->length());
         ldout(oc->cct, 20) << "map_read gap " << *n << dendl;
         continue;    // more?
-      }
-      else 
+      } else {
         assert(0);
+      }
     }
   }
-  return(0);
+  return 0;
 }
 
 /*
@@ -264,33 +255,16 @@ ObjectCacher::BufferHead *ObjectCacher::Object::map_write(OSDWrite *wr)
   for (vector<ObjectExtent>::iterator ex_it = wr->extents.begin();
        ex_it != wr->extents.end();
        ex_it++) {
-    
+
     if (ex_it->oid != oid.oid) continue;
-    
+
     ldout(oc->cct, 10) << "map_write oex " << ex_it->oid
-             << " " << ex_it->offset << "~" << ex_it->length << dendl;
-    
-    map<loff_t, BufferHead*>::iterator p = data.lower_bound(ex_it->offset);
-    // p->first >= start
-    
+		       << " " << ex_it->offset << "~" << ex_it->length << dendl;
+
     loff_t cur = ex_it->offset;
     loff_t left = ex_it->length;
-    
-    if (p != data.begin() && 
-        (p == data.end() || p->first > cur)) {
-      p--;     // might overlap or butt up!
 
-      /*// dirty and butts up?
-      if (p->first + p->second->length() == cur &&
-          p->second->is_dirty()) {
-        ldout(oc->cct, 10) << "map_write will append to tail of " << *p->second << dendl;
-        final = p->second;
-      }
-      */
-      if (p->first + p->second->length() <= cur) 
-        p++;   // doesn't overlap.
-    }    
-    
+    map<loff_t, BufferHead*>::iterator p = data_lower_bound(ex_it->offset);
     while (left > 0) {
       loff_t max = left;
 
@@ -346,8 +320,9 @@ ObjectCacher::BufferHead *ObjectCacher::Object::map_write(OSDWrite *wr)
 	    p--;  // move iterator back to final
 	    assert(p->second == final);
             merge_left(final, bh);
-	  } else
+	  } else {
             final = bh;
+	  }
         }
         
         // keep going.
@@ -412,14 +387,7 @@ void ObjectCacher::Object::discard(loff_t off, loff_t len)
 {
   ldout(oc->cct, 10) << "discard " << *this << " " << off << "~" << len << dendl;
 
-  map<loff_t, BufferHead*>::iterator p = data.lower_bound(off);
-  if (p != data.begin() &&
-      (p == data.end() || p->first > off)) {
-    p--;     // might overlap!
-    if (p->first + p->second->length() <= off)
-      p++;   // doesn't overlap.
-  }
-
+  map<loff_t, BufferHead*>::iterator p = data_lower_bound(off);
   while (p != data.end()) {
     BufferHead *bh = p->second;
     if (bh->start() >= off + len)
@@ -452,14 +420,17 @@ void ObjectCacher::Object::discard(loff_t off, loff_t len)
 
 ObjectCacher::ObjectCacher(CephContext *cct_, string name, WritebackHandler& wb, Mutex& l,
 			   flush_set_callback_t flush_callback,
-			   void *flush_callback_arg)
+			   void *flush_callback_arg,
+			   uint64_t max_size, uint64_t max_dirty, uint64_t target_dirty, double max_dirty_age)
   : perfcounter(NULL),
     cct(cct_), writeback_handler(wb), name(name), lock(l),
+    max_dirty(max_dirty), target_dirty(target_dirty), max_size(max_size),
     flush_set_callback(flush_callback), flush_set_callback_arg(flush_callback_arg),
     flusher_stop(false), flusher_thread(this),
-    stat_waiter(0),
-    stat_clean(0), stat_dirty(0), stat_rx(0), stat_tx(0), stat_missing(0)
+    stat_clean(0), stat_dirty(0), stat_rx(0), stat_tx(0), stat_missing(0),
+    stat_error(0), stat_dirty_waiting(0)
 {
+  this->max_dirty_age.set_from_double(max_dirty_age);
   perf_start();
 }
 
@@ -566,10 +537,6 @@ void ObjectCacher::bh_read_finish(int64_t poolid, sobject_t oid, loff_t start,
 		<< " returned " << r
 		<< dendl;
 
-  if (r < 0) {
-    // TODO: fix bad read case
-  }
-
   if (bl.length() < length) {
     bufferptr bp(length - bl.length());
     bp.zero();
@@ -613,7 +580,14 @@ void ObjectCacher::bh_read_finish(int64_t poolid, sobject_t oid, loff_t start,
       bh->bl.substr_of(bl,
                        opos-bh->start(),
                        bh->length());
-      mark_clean(bh);
+
+      if (r < 0 && r != -ENOENT) {
+	mark_error(bh);
+	bh->error = r;
+      } else {
+	mark_clean(bh);
+      }
+
       ldout(cct, 10) << "bh_read_finish read " << *bh << dendl;
       
       opos = bh->end();
@@ -627,7 +601,7 @@ void ObjectCacher::bh_read_finish(int64_t poolid, sobject_t oid, loff_t start,
            p++)
         ls.splice(ls.end(), p->second);
       bh->waitfor_read.clear();
-      finish_contexts(cct, ls);
+      finish_contexts(cct, ls, bh->error);
 
       // clean up?
       ob->try_merge_bh(bh);
@@ -735,10 +709,8 @@ void ObjectCacher::bh_write_commit(int64_t poolid, sobject_t oid, loff_t start,
 		<< oid 
 		<< " tid " << tid
 		<< " " << start << "~" << length
+		<< " returned " << r
 		<< dendl;
-  if (r < 0) {
-    // TODO: handle write error
-  }
 
   if (objects[poolid].count(oid) == 0) {
     ldout(cct, 7) << "bh_write_commit no object cache" << dendl;
@@ -772,12 +744,19 @@ void ObjectCacher::bh_write_commit(int64_t poolid, sobject_t oid, loff_t start,
         ldout(cct, 10) << "bh_write_commit newer tid on " << *bh << dendl;
         continue;
       }
-      
-      // ok!  mark bh clean.
-      mark_clean(bh);
-      ldout(cct, 10) << "bh_write_commit clean " << *bh << dendl;
+
+      if (r >= 0) {
+	// ok!  mark bh clean and error-free
+	mark_clean(bh);
+	ldout(cct, 10) << "bh_write_commit clean " << *bh << dendl;
+      } else {
+	mark_dirty(bh);
+	ldout(cct, 10) << "bh_write_commit marking dirty again due to error "
+		       << *bh << " r = " << r << " " << cpp_strerror(-r)
+		       << dendl;
+      }
     }
-    
+
     // update last_commit.
     assert(ob->last_commit_tid < tid);
     ob->last_commit_tid = tid;
@@ -787,7 +766,7 @@ void ObjectCacher::bh_write_commit(int64_t poolid, sobject_t oid, loff_t start,
       list<Context*> ls;
       ls.splice(ls.begin(), ob->waitfor_commit[tid]);
       ob->waitfor_commit.erase(tid);
-      finish_contexts(cct, ls);
+      finish_contexts(cct, ls, r);
     }
 
     // is the entire object set now clean and fully committed?
@@ -806,7 +785,6 @@ void ObjectCacher::bh_write_commit(int64_t poolid, sobject_t oid, loff_t start,
 void ObjectCacher::flush(loff_t amount)
 {
   utime_t cutoff = ceph_clock_now(cct);
-  //cutoff.sec_ref() -= cct->_conf->client_oc_max_dirty_age;
 
   ldout(cct, 10) << "flush " << amount << dendl;
   
@@ -830,7 +808,7 @@ void ObjectCacher::flush(loff_t amount)
 void ObjectCacher::trim(loff_t max)
 {
   if (max < 0) 
-    max = cct->_conf->client_oc_size;
+    max = max_size;
   
   ldout(cct, 10) << "trim  start: max " << max 
            << "  clean " << get_stat_clean()
@@ -886,9 +864,16 @@ bool ObjectCacher::is_cached(ObjectSet *oset, vector<ObjectExtent>& extents, sna
  * returns 0 if doing async read
  */
 int ObjectCacher::readx(OSDRead *rd, ObjectSet *oset, Context *onfinish)
+{
+  return _readx(rd, oset, onfinish, true);
+}
+
+int ObjectCacher::_readx(OSDRead *rd, ObjectSet *oset, Context *onfinish,
+			 bool external_call)
 {
   assert(lock.is_locked());
   bool success = true;
+  int error = 0;
   list<BufferHead*> hit_ls;
   uint64_t bytes_in_cache = 0;
   uint64_t bytes_not_in_cache = 0;
@@ -907,8 +892,17 @@ int ObjectCacher::readx(OSDRead *rd, ObjectSet *oset, Context *onfinish)
     Object *o = get_object(soid, oset, ex_it->oloc);
     
     // map extent into bufferheads
-    map<loff_t, BufferHead*> hits, missing, rx;
-    o->map_read(rd, hits, missing, rx);
+    map<loff_t, BufferHead*> hits, missing, rx, errors;
+    o->map_read(rd, hits, missing, rx, errors);
+    if (external_call) {
+      // retry reading error buffers
+      missing.insert(errors.begin(), errors.end());
+    } else {
+      // some reads had errors, fail later so completions
+      // are cleaned up up properly
+      // TODO: make read path not call _readx for every completion
+      hits.insert(errors.begin(), errors.end());
+    }
     
     if (!missing.empty() || !rx.empty()) {
       // read missing
@@ -946,6 +940,8 @@ int ObjectCacher::readx(OSDRead *rd, ObjectSet *oset, Context *onfinish)
            bh_it != hits.end();
            bh_it++) {
 	ldout(cct, 10) << "readx hit bh " << *bh_it->second << dendl;
+	if (bh_it->second->is_error() && bh_it->second->error)
+	  error = bh_it->second->error;
         hit_ls.push_back(bh_it->second);
         bytes_in_cache += bh_it->second->length();
       }
@@ -1004,7 +1000,7 @@ int ObjectCacher::readx(OSDRead *rd, ObjectSet *oset, Context *onfinish)
     touch_bh(*bhit);
   
   if (!success) {
-    if (perfcounter) {
+    if (perfcounter && external_call) {
       perfcounter->inc(l_objectcacher_data_read, total_bytes_read);
       perfcounter->inc(l_objectcacher_cache_bytes_hit, bytes_in_cache);
       perfcounter->inc(l_objectcacher_cache_bytes_miss, bytes_not_in_cache);
@@ -1012,7 +1008,7 @@ int ObjectCacher::readx(OSDRead *rd, ObjectSet *oset, Context *onfinish)
     }
     return 0;  // wait!
   }
-  if (perfcounter) {
+  if (perfcounter && external_call) {
     perfcounter->inc(l_objectcacher_data_read, total_bytes_read);
     perfcounter->inc(l_objectcacher_cache_bytes_hit, bytes_in_cache);
     perfcounter->inc(l_objectcacher_cache_ops_hit);
@@ -1024,7 +1020,7 @@ int ObjectCacher::readx(OSDRead *rd, ObjectSet *oset, Context *onfinish)
   
   // ok, assemble into result buffer.
   uint64_t pos = 0;
-  if (rd->bl) {
+  if (rd->bl && !error) {
     rd->bl->clear();
     for (map<uint64_t,bufferlist>::iterator i = stripe_map.begin();
 	 i != stripe_map.end();
@@ -1044,12 +1040,14 @@ int ObjectCacher::readx(OSDRead *rd, ObjectSet *oset, Context *onfinish)
   delete rd;
 
   trim();
-  
-  return pos;
+
+  assert(pos <= (uint64_t) INT_MAX);
+
+  return error ? error : pos;
 }
 
 
-int ObjectCacher::writex(OSDWrite *wr, ObjectSet *oset)
+int ObjectCacher::writex(OSDWrite *wr, ObjectSet *oset, Mutex& wait_on_lock)
 {
   assert(lock.is_locked());
   utime_t now = ceph_clock_now(cct);
@@ -1116,39 +1114,58 @@ int ObjectCacher::writex(OSDWrite *wr, ObjectSet *oset)
     }
   }
 
+  int r = _wait_for_write(wr, bytes_written, oset, wait_on_lock);
+
   delete wr;
 
   //verify_stats();
   trim();
-  return 0;
+  return r;
 }
  
 
 // blocking wait for write.
-bool ObjectCacher::wait_for_write(uint64_t len, Mutex& lock)
+int ObjectCacher::_wait_for_write(OSDWrite *wr, uint64_t len, ObjectSet *oset, Mutex& lock)
 {
   int blocked = 0;
-  const md_config_t *conf = cct->_conf;
   utime_t start = ceph_clock_now(cct);
+  int ret = 0;
 
-  // wait for writeback?
-  while (get_stat_dirty() + get_stat_tx() >= conf->client_oc_max_dirty) {
-    ldout(cct, 10) << "wait_for_write waiting on " << len << ", dirty|tx " 
-	     << (get_stat_dirty() + get_stat_tx()) 
-	     << " >= " << conf->client_oc_max_dirty 
-	     << dendl;
-    flusher_cond.Signal();
-    stat_waiter++;
-    stat_cond.Wait(lock);
-    stat_waiter--;
-    blocked++;
-    ldout(cct, 10) << "wait_for_write woke up" << dendl;
+  if (max_dirty > 0) {
+    // wait for writeback?
+    //  - wait for dirty and tx bytes (relative to the max_dirty threshold)
+    //  - do not wait for bytes other waiters are waiting on.  this means that
+    //    threads do not wait for each other.  this effectively allows the cache size
+    //    to balloon proportional to the data that is in flight.
+    while (get_stat_dirty() + get_stat_tx() >= max_dirty + get_stat_dirty_waiting()) {
+      ldout(cct, 10) << "wait_for_write waiting on " << len << ", dirty|tx " 
+		     << (get_stat_dirty() + get_stat_tx()) 
+		     << " >= max " << max_dirty << " + dirty_waiting " << get_stat_dirty_waiting()
+		     << dendl;
+      flusher_cond.Signal();
+      stat_dirty_waiting += len;
+      stat_cond.Wait(lock);
+      stat_dirty_waiting -= len;
+      blocked++;
+      ldout(cct, 10) << "wait_for_write woke up" << dendl;
+    }
+  } else {
+    // write-thru!  flush what we just wrote.
+    Cond cond;
+    bool done;
+    C_Cond *fin = new C_Cond(&cond, &done, &ret);
+    bool flushed = flush_set(oset, wr->extents, fin);
+    assert(!flushed);   // we just dirtied it, and didn't drop our lock!
+    ldout(cct, 10) << "wait_for_write waiting on write-thru of " << len << " bytes" << dendl;
+    while (!done)
+      cond.Wait(lock);
+    ldout(cct, 10) << "wait_for_write woke up, ret " << ret << dendl;
   }
 
   // start writeback anyway?
-  if (get_stat_dirty() > conf->client_oc_target_dirty) {
+  if (get_stat_dirty() > target_dirty) {
     ldout(cct, 10) << "wait_for_write " << get_stat_dirty() << " > target "
-	     << conf->client_oc_target_dirty << ", nudging flusher" << dendl;
+		   << target_dirty << ", nudging flusher" << dendl;
     flusher_cond.Signal();
   }
   if (blocked && perfcounter) {
@@ -1157,48 +1174,46 @@ bool ObjectCacher::wait_for_write(uint64_t len, Mutex& lock)
     utime_t blocked = ceph_clock_now(cct) - start;
     perfcounter->finc(l_objectcacher_write_time_blocked, (double) blocked);
   }
-  return blocked;
+  return ret;
 }
 
 void ObjectCacher::flusher_entry()
 {
-  const md_config_t *conf = cct->_conf;
   ldout(cct, 10) << "flusher start" << dendl;
   lock.Lock();
   while (!flusher_stop) {
-    while (!flusher_stop) {
-      loff_t all = get_stat_tx() + get_stat_rx() + get_stat_clean() + get_stat_dirty();
-      ldout(cct, 11) << "flusher "
-               << all << " / " << conf->client_oc_size << ":  "
-               << get_stat_tx() << " tx, "
-               << get_stat_rx() << " rx, "
-               << get_stat_clean() << " clean, "
-               << get_stat_dirty() << " dirty ("
-	       << conf->client_oc_target_dirty << " target, "
-	       << conf->client_oc_max_dirty << " max)"
-               << dendl;
-      if (get_stat_dirty() > conf->client_oc_target_dirty) {
-        // flush some dirty pages
-        ldout(cct, 10) << "flusher " 
-                 << get_stat_dirty() << " dirty > target "
-		 << conf->client_oc_target_dirty
-                 << ", flushing some dirty bhs" << dendl;
-        flush(get_stat_dirty() - conf->client_oc_target_dirty);
-      }
-      else {
-        // check tail of lru for old dirty items
-        utime_t cutoff = ceph_clock_now(cct);
-        cutoff.sec_ref()--;
-        BufferHead *bh = 0;
-        while ((bh = (BufferHead*)lru_dirty.lru_get_next_expire()) != 0 &&
-               bh->last_write < cutoff) {
-          ldout(cct, 10) << "flusher flushing aged dirty bh " << *bh << dendl;
-          bh_write(bh);
-        }
-        break;
+    loff_t all = get_stat_tx() + get_stat_rx() + get_stat_clean() + get_stat_dirty();
+    ldout(cct, 11) << "flusher "
+		   << all << " / " << max_size << ":  "
+		   << get_stat_tx() << " tx, "
+		   << get_stat_rx() << " rx, "
+		   << get_stat_clean() << " clean, "
+		   << get_stat_dirty() << " dirty ("
+		   << target_dirty << " target, "
+		   << max_dirty << " max)"
+		   << dendl;
+    loff_t actual = get_stat_dirty() + get_stat_dirty_waiting();
+    if (actual > target_dirty) {
+      // flush some dirty pages
+      ldout(cct, 10) << "flusher " 
+		     << get_stat_dirty() << " dirty + " << get_stat_dirty_waiting()
+		     << " dirty_waiting > target "
+		     << target_dirty
+		     << ", flushing some dirty bhs" << dendl;
+      flush(actual - target_dirty);
+    } else {
+      // check tail of lru for old dirty items
+      utime_t cutoff = ceph_clock_now(cct);
+      cutoff -= max_dirty_age;
+      BufferHead *bh = 0;
+      while ((bh = (BufferHead*)lru_dirty.lru_get_next_expire()) != 0 &&
+	     bh->last_write < cutoff) {
+	ldout(cct, 10) << "flusher flushing aged dirty bh " << *bh << dendl;
+	bh_write(bh);
       }
     }
-    if (flusher_stop) break;
+    if (flusher_stop)
+      break;
     flusher_cond.WaitInterval(cct, lock, utime_t(1,0));
   }
   lock.Unlock();
@@ -1333,7 +1348,7 @@ void ObjectCacher::wrunlock(Object *o)
     return;
   }
 
-  flush(o);  // flush first
+  flush(o, 0, 0);  // flush first
 
   int op = 0;
   if (o->rdlock_ref > 0) {
@@ -1418,19 +1433,24 @@ void ObjectCacher::purge(Object *ob)
 // flush.  non-blocking.  no callback.
 // true if clean, already flushed.  
 // false if we wrote something.
-bool ObjectCacher::flush(Object *ob)
+// be sloppy about the ranges and flush any buffer it touches
+bool ObjectCacher::flush(Object *ob, loff_t offset, loff_t length)
 {
   bool clean = true;
-  for (map<loff_t,BufferHead*>::iterator p = ob->data.begin();
-       p != ob->data.end();
-       p++) {
+  ldout(cct, 10) << "flush " << *ob << " " << offset << "~" << length << dendl;
+  for (map<loff_t,BufferHead*>::iterator p = ob->data_lower_bound(offset); p != ob->data.end(); p++) {
     BufferHead *bh = p->second;
+    ldout(cct, 20) << "flush  " << *bh << dendl;
+    if (length && bh->start() > offset+length) {
+      break;
+    }
     if (bh->is_tx()) {
       clean = false;
       continue;
     }
-    if (!bh->is_dirty()) continue;
-    
+    if (!bh->is_dirty()) {
+      continue;
+    }
     bh_write(bh);
     clean = false;
   }
@@ -1443,6 +1463,7 @@ bool ObjectCacher::flush_set(ObjectSet *oset, Context *onfinish)
 {
   if (oset->objects.empty()) {
     ldout(cct, 10) << "flush_set on " << oset << " dne" << dendl;
+    delete onfinish;
     return true;
   }
 
@@ -1456,7 +1477,7 @@ bool ObjectCacher::flush_set(ObjectSet *oset, Context *onfinish)
        !i.end(); ++i) {
     Object *ob = *i;
 
-    if (!flush(ob)) {
+    if (!flush(ob, 0, 0)) {
       // we'll need to gather...
       safe = false;
 
@@ -1473,6 +1494,55 @@ bool ObjectCacher::flush_set(ObjectSet *oset, Context *onfinish)
   
   if (safe) {
     ldout(cct, 10) << "flush_set " << oset << " has no dirty|tx bhs" << dendl;
+    delete onfinish;
+    return true;
+  }
+  return false;
+}
+
+// flush.  non-blocking, takes callback.
+// returns true if already flushed
+bool ObjectCacher::flush_set(ObjectSet *oset, vector<ObjectExtent>& exv, Context *onfinish)
+{
+  if (oset->objects.empty()) {
+    ldout(cct, 10) << "flush_set on " << oset << " dne" << dendl;
+    delete onfinish;
+    return true;
+  }
+
+  ldout(cct, 10) << "flush_set " << oset << " on " << exv.size() << " ObjectExtents" << dendl;
+
+  // we'll need to wait for all objects to flush!
+  C_GatherBuilder gather(cct, onfinish);
+
+  bool safe = true;
+  for (vector<ObjectExtent>::iterator p = exv.begin();
+       p != exv.end();
+       ++p) {
+    ObjectExtent &ex = *p;
+    sobject_t soid(ex.oid, CEPH_NOSNAP);
+    if (objects[oset->poolid].count(soid) == 0)
+      continue;
+    Object *ob = objects[oset->poolid][soid];
+
+    ldout(cct, 20) << "flush_set " << oset << " ex " << ex << " ob " << soid << " " << ob << dendl;
+
+    if (!flush(ob, ex.offset, ex.length)) {
+      // we'll need to gather...
+      safe = false;
+
+      ldout(cct, 10) << "flush_set " << oset << " will wait for ack tid " 
+		     << ob->last_write_tid << " on " << *ob << dendl;
+      if (onfinish != NULL)
+        ob->waitfor_commit[ob->last_write_tid].push_back(gather.new_sub());
+    }
+  }
+  if (onfinish != NULL)
+    gather.activate();
+  
+  if (safe) {
+    ldout(cct, 10) << "flush_set " << oset << " has no dirty|tx bhs" << dendl;
+    delete onfinish;
     return true;
   }
   return false;
@@ -1686,7 +1756,7 @@ void ObjectCacher::verify_stats() const
 {
   ldout(cct, 10) << "verify_stats" << dendl;
 
-  loff_t clean = 0, dirty = 0, rx = 0, tx = 0, missing = 0;
+  loff_t clean = 0, dirty = 0, rx = 0, tx = 0, missing = 0, error = 0;
   for (vector<hash_map<sobject_t, Object*> >::const_iterator i = objects.begin();
       i != objects.end();
       ++i) {
@@ -1714,6 +1784,9 @@ void ObjectCacher::verify_stats() const
         case BufferHead::STATE_RX:
           rx += bh->length();
           break;
+	case BufferHead::STATE_ERROR:
+	  error += bh->length();
+	  break;
         default:
           assert(0);
         }
@@ -1726,12 +1799,14 @@ void ObjectCacher::verify_stats() const
 	   << " tx " << tx
 	   << " dirty " << dirty
 	   << " missing " << missing
+	   << " error " << error
 	   << dendl;
   assert(clean == stat_clean);
   assert(rx == stat_rx);
   assert(tx == stat_tx);
   assert(dirty == stat_dirty);
   assert(missing == stat_missing);
+  assert(error == stat_error);
 }
 
 void ObjectCacher::bh_stat_add(BufferHead *bh)
@@ -1756,8 +1831,14 @@ void ObjectCacher::bh_stat_add(BufferHead *bh)
   case BufferHead::STATE_RX:
     stat_rx += bh->length();
     break;
+  case BufferHead::STATE_ERROR:
+    stat_error += bh->length();
+    break;
+  default:
+    assert(0 == "bh_stat_add: invalid bufferhead state");
   }
-  if (stat_waiter) stat_cond.Signal();
+  if (get_stat_dirty_waiting() > 0)
+    stat_cond.Signal();
 }
 
 void ObjectCacher::bh_stat_sub(BufferHead *bh)
@@ -1782,6 +1863,11 @@ void ObjectCacher::bh_stat_sub(BufferHead *bh)
   case BufferHead::STATE_RX:
     stat_rx -= bh->length();
     break;
+  case BufferHead::STATE_ERROR:
+    stat_error -= bh->length();
+    break;
+  default:
+    assert(0 == "bh_stat_sub: invalid bufferhead state");
   }
 }
 
@@ -1798,6 +1884,9 @@ void ObjectCacher::bh_set_state(BufferHead *bh, int s)
     lru_rest.lru_insert_top(bh);
     dirty_bh.erase(bh);
   }
+  if (s != BufferHead::STATE_ERROR && bh->get_state() == BufferHead::STATE_ERROR) {
+    bh->error = 0;
+  }
 
   // set state
   bh_stat_sub(bh);
