@@ -14,14 +14,14 @@ using namespace std;
 
 #include "common/armor.h"
 #include "rgw_user.h"
-#include "rgw_access.h"
+#include "rgw_rados.h"
 #include "rgw_acl.h"
 #include "rgw_acl_s3.h"
 #include "rgw_log.h"
 #include "rgw_formats.h"
 #include "auth/Crypto.h"
 
-#define DOUT_SUBSYS rgw
+#define dout_subsys ceph_subsys_rgw
 
 #define SECRET_KEY_LEN 40
 #define PUBLIC_ID_LEN 20
@@ -296,10 +296,10 @@ static void show_user_info(RGWUserInfo& info, Formatter *formatter)
 
   formatter->open_object_section("user_info");
 
-  formatter->dump_string("user_id", info.user_id.c_str());
+  formatter->dump_string("user_id", info.user_id);
   formatter->dump_int("rados_uid", info.auid);
-  formatter->dump_string("display_name", info.display_name.c_str());
-  formatter->dump_string("email", info.user_email.c_str());
+  formatter->dump_string("display_name", info.display_name);
+  formatter->dump_string("email", info.user_email);
   formatter->dump_int("suspended", (int)info.suspended);
 
   // subusers
@@ -473,12 +473,12 @@ int bucket_stats(rgw_bucket& bucket, Formatter *formatter)
   }
   map<RGWObjCategory, RGWBucketStats>::iterator iter;
   formatter->open_object_section("stats");
-  formatter->dump_string("bucket", bucket.name.c_str());
-  formatter->dump_string("pool", bucket.pool.c_str());
+  formatter->dump_string("bucket", bucket.name);
+  formatter->dump_string("pool", bucket.pool);
   
-  formatter->dump_int("id", bucket.bucket_id);
-  formatter->dump_string("marker", bucket.marker.c_str());
-  formatter->dump_string("owner", bucket_info.owner.c_str());
+  formatter->dump_string("id", bucket.bucket_id);
+  formatter->dump_string("marker", bucket.marker);
+  formatter->dump_string("owner", bucket_info.owner);
   formatter->open_object_section("usage");
   for (iter = stats.begin(); iter != stats.end(); ++iter) {
     RGWBucketStats& s = iter->second;
@@ -518,7 +518,7 @@ int main(int argc, char **argv)
   uint32_t perm_mask = 0;
   uint64_t auid = -1;
   RGWUserInfo info;
-  RGWAccess *store;
+  RGWRados *store;
   int opt_cmd = OPT_NO_CMD;
   bool need_more;
   int gen_secret = false;
@@ -528,7 +528,7 @@ int main(int argc, char **argv)
   char secret_key_buf[SECRET_KEY_LEN + 1];
   char public_id_buf[PUBLIC_ID_LEN + 1];
   bool user_modify_op;
-  int64_t bucket_id = -1;
+  string bucket_id;
   Formatter *formatter = NULL;
   int purge_data = false;
   RGWBucketInfo bucket_info;
@@ -598,14 +598,10 @@ int main(int argc, char **argv)
     } else if (ceph_argparse_witharg(args, i, &val, "--access", (char*)NULL)) {
       access = val;
       perm_mask = str_to_perm(access.c_str());
-    } else if (ceph_argparse_withlonglong(args, i, &tmp, &errs, "--bucket-id", (char*)NULL)) {
-      if (!errs.str().empty()) {
-	cerr << errs.str() << std::endl;
-	exit(EXIT_FAILURE);
-      }
-      bucket_id = tmp;
-      if (bucket_id < 0) {
-        cerr << "bad bucket-id: " << bucket_id << std::endl;
+    } else if (ceph_argparse_witharg(args, i, &val, "--bucket-id", (char*)NULL)) {
+      bucket_id = val;
+      if (bucket_id.empty()) {
+        cerr << "bad bucket-id" << std::endl;
         return usage();
       }
     } else if (ceph_argparse_witharg(args, i, &val, "--format", (char*)NULL)) {
@@ -683,7 +679,7 @@ int main(int argc, char **argv)
                     opt_cmd == OPT_KEY_CREATE || opt_cmd == OPT_KEY_RM || opt_cmd == OPT_USER_RM);
 
   RGWStoreManager store_manager;
-  store = store_manager.init("rados", g_ceph_context);
+  store = store_manager.init(g_ceph_context);
   if (!store) {
     cerr << "couldn't init storage provider" << std::endl;
     return 5; //EIO
@@ -727,8 +723,14 @@ int main(int argc, char **argv)
 
     if (opt_cmd == OPT_USER_CREATE) {
       if (found) {
-        cerr << "error: user already exists" << std::endl;
-        return 1;
+        if (info.display_name.compare(display_name) != 0 ||
+            info.user_email.compare(user_email) != 0) {
+          cerr << "error: user already exists with different display_name/email" << std::endl;
+          return 1;
+        }
+        /* turn into OPT_USER_MODIFY */
+        opt_cmd = OPT_USER_MODIFY;
+        user_modify_op = true;
       }
     } else if (!found) {
       cerr << "error reading user info, aborting" << std::endl;
@@ -736,20 +738,30 @@ int main(int argc, char **argv)
     }
   }
 
+  bool subuser_found = false;
+
+  if (!subuser.empty()) {
+    map<string, RGWSubUser>::iterator iter = info.subusers.find(subuser);
+    subuser_found = (iter != info.subusers.end());
+
+    if (!subuser_found && opt_cmd != OPT_SUBUSER_CREATE && opt_cmd != OPT_USER_CREATE) {
+      cerr << "subuser specified but was not found, aborting" << std::endl;
+      return 1;
+    }
+  }
+
   if (opt_cmd == OPT_SUBUSER_CREATE || opt_cmd == OPT_SUBUSER_MODIFY ||
       opt_cmd == OPT_SUBUSER_RM) {
     if (subuser.empty()) {
       cerr << "subuser creation was requires specifying subuser name" << std::endl;
       return 1;
     }
-    map<string, RGWSubUser>::iterator iter = info.subusers.find(subuser);
-    bool found = (iter != info.subusers.end());
     if (opt_cmd == OPT_SUBUSER_CREATE) {
-      if (found) {
+      if (subuser_found) {
         cerr << "error: subuser already exists" << std::endl;
         return 1;
       }
-    } else if (!found) {
+    } else if (!subuser_found) {
       cerr << "error: subuser doesn't exist" << std::endl;
       return 1;
     }
@@ -767,7 +779,7 @@ int main(int argc, char **argv)
     }
 
     if ((secret_key.empty() && implicit_gen_secret) || gen_secret) {
-      ret = gen_rand_base64(secret_key_buf, sizeof(secret_key_buf));
+      ret = gen_rand_base64(g_ceph_context, secret_key_buf, sizeof(secret_key_buf));
       if (ret < 0) {
         cerr << "aborting" << std::endl;
         return 1;
@@ -778,7 +790,7 @@ int main(int argc, char **argv)
       RGWUserInfo duplicate_check;
       string duplicate_check_id;
       do {
-	ret = gen_rand_alphanumeric_upper(public_id_buf, sizeof(public_id_buf));
+	ret = gen_rand_alphanumeric_upper(g_ceph_context, public_id_buf, sizeof(public_id_buf));
 	if (ret < 0) {
 	  cerr << "aborting" << std::endl;
 	  return 1;
@@ -793,13 +805,9 @@ int main(int argc, char **argv)
   map<string, RGWSubUser>::iterator uiter;
   RGWUserInfo old_info = info;
 
-  if ((!bucket_name.empty()) || bucket_id >= 0) {
-    if (bucket_id >= 0) {
-      char bucket_char[16];
-      snprintf(bucket_char, sizeof(bucket_char), ".%lld",
-               (long long unsigned)bucket_id);
-      string bucket_string(bucket_char);
-      int ret = rgwstore->get_bucket_info(NULL, bucket_string, bucket_info);
+  if ((!bucket_name.empty()) || !bucket_id.empty()) {
+    if (!bucket_id.empty()) {
+      int ret = rgwstore->get_bucket_info(NULL, bucket_id, bucket_info);
 
       if (ret < 0) {
         cerr << "could not retrieve bucket info for bucket_id=" << bucket_id << std::endl;
@@ -942,7 +950,7 @@ int main(int argc, char **argv)
     rgw_obj obj(bucket, object);
     int ret = store->get_attr(NULL, obj, RGW_ATTR_ACL, bl);
 
-    RGWAccessControlPolicy_S3 policy;
+    RGWAccessControlPolicy_S3 policy(g_ceph_context);
     if (ret >= 0) {
       bufferlist::iterator iter = bl.begin();
       try {
@@ -1086,7 +1094,7 @@ int main(int argc, char **argv)
   }
 
   if (opt_cmd == OPT_LOG_SHOW || opt_cmd == OPT_LOG_RM) {
-    if (object.empty() && (date.empty() || bucket_name.empty() || bucket_id < 0)) {
+    if (object.empty() && (date.empty() || bucket_name.empty() || bucket_id.empty())) {
       cerr << "object or (at least one of date, bucket, bucket-id) were not specified" << std::endl;
       return usage();
     }
@@ -1095,11 +1103,9 @@ int main(int argc, char **argv)
     if (!object.empty()) {
       oid = object;
     } else {
-      char buf[16];
-      snprintf(buf, sizeof(buf), "%lld", (unsigned long long)bucket_id);
       oid = date;
       oid += "-";
-      oid += buf;
+      oid += bucket_id;
       oid += "-";
       oid += string(bucket.name);
     }
@@ -1124,7 +1130,7 @@ int main(int argc, char **argv)
 	cerr << "error reading log " << oid << ": " << cpp_strerror(-r) << std::endl;
 	return -r;
       }
-      formatter->dump_int("bucket_id", entry.bucket_id);
+      formatter->dump_string("bucket_id", entry.bucket_id);
       formatter->dump_string("bucket_owner", entry.bucket_owner);
       formatter->dump_string("bucket", entry.bucket);
 
@@ -1150,23 +1156,23 @@ int main(int argc, char **argv)
 
         if (show_log_entries) {
 	  formatter->open_object_section("log_entry");
-	  formatter->dump_string("bucket", entry.bucket.c_str());
+	  formatter->dump_string("bucket", entry.bucket);
 	  entry.time.gmtime(formatter->dump_stream("time"));      // UTC
 	  entry.time.localtime(formatter->dump_stream("time_local"));
-	  formatter->dump_string("remote_addr", entry.remote_addr.c_str());
+	  formatter->dump_string("remote_addr", entry.remote_addr);
 	  if (entry.object_owner.length())
-	    formatter->dump_string("object_owner", entry.object_owner.c_str());
-	  formatter->dump_string("user", entry.user.c_str());
-	  formatter->dump_string("operation", entry.op.c_str());
-	  formatter->dump_string("uri", entry.uri.c_str());
-	  formatter->dump_string("http_status", entry.http_status.c_str());
-	  formatter->dump_string("error_code", entry.error_code.c_str());
+	    formatter->dump_string("object_owner", entry.object_owner);
+	  formatter->dump_string("user", entry.user);
+	  formatter->dump_string("operation", entry.op);
+	  formatter->dump_string("uri", entry.uri);
+	  formatter->dump_string("http_status", entry.http_status);
+	  formatter->dump_string("error_code", entry.error_code);
 	  formatter->dump_int("bytes_sent", entry.bytes_sent);
 	  formatter->dump_int("bytes_received", entry.bytes_received);
 	  formatter->dump_int("object_size", entry.obj_size);
 	  formatter->dump_int("total_time", total_time);
-	  formatter->dump_string("user_agent",  entry.user_agent.c_str());
-	  formatter->dump_string("referrer",  entry.referrer.c_str());
+	  formatter->dump_string("user_agent",  entry.user_agent);
+	  formatter->dump_string("referrer",  entry.referrer);
 	  formatter->close_section();
 	  formatter->flush(cout);
         }
@@ -1203,6 +1209,14 @@ next:
   }
   
   if (opt_cmd == OPT_USER_RM) {
+    RGWUserBuckets buckets;
+    if (rgw_read_user_buckets(user_id, buckets, false) >= 0) {
+      map<string, RGWBucketEnt>& m = buckets.get_buckets();
+      if (m.size() > 0) {
+        cerr << "user bucket list not empty, can't remove user" << std::endl;
+        return 1;
+      }
+    }
     rgw_delete_user(info);
   }
   
@@ -1249,7 +1263,7 @@ next:
   }
 
   if (opt_cmd == OPT_BUCKET_STATS) {
-    if (bucket_name.empty() && bucket_id < 0 && user_id.empty()) {
+    if (bucket_name.empty() && bucket_id.empty() && user_id.empty()) {
       cerr << "either bucket or bucket-id or uid needs to be specified" << std::endl;
       return usage();
     }
